#include <stdio.h>
#include "cmmd.h"

CMMD_mcb CMMD_send_async(dest_node, tag, buffer, buffer_desc, handler,
                         handler_arg)
int dest_node;
int tag;
void *buffer;
int buffer_desc;
void (*handler)();
void *handler_arg;
{
   CMMD_mcb mcb;
   MPI_Request request;

   if ((mcb = (CMMD_mcb) malloc(sizeof(struct mcb))) == NULL)
   {
      printf("CMMD_send_async(): Out of memory!\n");
      exit(1);
   }

   mcb -> node = machine_info.my_address;
   mcb -> source = machine_info.my_address;
   mcb -> dest = dest_node;
   mcb -> tag = tag;
   mcb -> bytes = buffer_desc;
   mcb -> buffer = buffer;

   MPE_Log_send(dest_node, tag, buffer_desc);
   MPE_Log_event(7,0,"Async. Send");
   MPI_Isend(buffer, buffer_desc, MPI_BYTE, dest_node, tag, MPI_COMM_WORLD,
             &(mcb -> request));
   MPE_Log_event(8,0,"Async. Send");

#ifdef DEBUG_CMMD_send_async
   printf("CMMD_send_async(): msgid = %d\n", msgid);
#endif

   mcb -> handler = handler;
   mcb -> handler_arg = handler_arg;

   if (handler != NULL)
      (*handler)(&mcb, mcb -> handler_arg);

   return(mcb);
}

CMMD_mcb CMMD_receive_async(source_node, tag, buffer, buffer_desc, handler,
                            handler_arg)
int source_node;
int tag;
void *buffer;
int buffer_desc;
void (*handler)();
void *handler_arg;
{
   CMMD_mcb mcb;

   if ((mcb = (CMMD_mcb) malloc(sizeof(struct mcb))) == NULL)
   {
      printf("CMMD_send_async(): Out of memory!\n");
      exit(1);
   }

   mcb -> node = machine_info.my_address;
   mcb -> dest = machine_info.my_address;
   mcb -> bytes = buffer_desc;
   mcb -> buffer = buffer;

#ifdef DEBUG_CMMD_receive_async
   printf("CMMD_receive_async(): before mpc_recv\n");
   printf("buffer = %lx, buffer_size = %d, source = %d, type = %d\n",
          buffer, buffer_desc, source, type);
#endif

   MPE_Log_event(9,0,"Async. Receive");
   MPI_Irecv(buffer, buffer_desc, MPI_BYTE, source_node, tag, MPI_COMM_WORLD,
             &(mcb -> request));
   MPE_Log_event(10,0,"Async. Receive");

#ifdef DEBUG_CMMD_receive_async
   printf("CMMD_receive_async(): after mpc_recv\n");
#endif

   mcb -> source = source_node;
   mcb -> tag = tag;
   mcb -> handler = handler;
   mcb -> handler_arg = handler_arg;

   *((int *) handler_arg) = ~tag;

#ifdef DEBUG_CMMD_receive_async
   printf("CMMD_receive_async(): msgid = %d\n", msgid);
#endif

   return(mcb);
}

void CMMD_msg_wait(mcb)
CMMD_mcb mcb;
{
   MPI_Status status;

   if (mcb != NULL)
   {
#ifdef DEBUG_CMMD_msg_wait
      printf("CMMD_msg_wait(): request = %lx\n", &(mdb -> request));
#endif
      MPE_Log_event(11,0,"Async. Wait");
      MPI_Wait(&(mcb -> request), &status);
      MPE_Log_event(12,0,"Async. Wait");
      MPE_Log_receive(status.MPI_SOURCE, status.MPI_TAG, mcb -> bytes);

      mcb -> source = status.MPI_SOURCE;
      mcb -> tag = status.MPI_TAG;

      if (mcb -> handler != NULL)
         (*(mcb -> handler))(&mcb, mcb -> handler_arg);
   }
}

